"""
This module is dedicated to the correlation matrix generation analysis function,
a data analysis function

"""

import sys
from pathlib import Path


def generate_csv(
    symmetric_corr_table, rna_info, data_load_sources, stringency, out_dir
):
    """
    Given a correlation table of RBPs binding to an RNA molecule of interest,
    generates and saves a CSV file containing the correlation values.
    :param symmetric_corr_table: A nested dictionary such that
        symmetric_corr_table[rbp1][rbp2] returns the binding correlation value
        between rbp1 and rbp2.
    :param rna_info: A dictionary containing 'official_name' as key, with a
        corresponding value that indicates the name of the RNA under
        investigation.
    :param data_load_sources: A list of data sources that were used, such as
        'postar', 'rbpdb', etc.
    :param stringency: numerical value representing the base stingency that was
        used in analyzing the correlation coefficient.
    :returns: a filepath to the saved CSV file.

    """
    rna = rna_info["official_name"]

    path_to_save = (
        Path(out_dir)
        / "csv"
        / f"{rna.lower()}-{'-'.join(data_load_sources)}-{stringency}.csv"
    )

    path_to_save.parent.mkdir(parents=True, exist_ok=True)
    path_to_save = str(path_to_save)

    with open(path_to_save, "w+") as csv_file:
        rbps = list(symmetric_corr_table.keys())
        csv_file.write("," + ",".join(rbps))
        csv_file.write("\n")
        for rbp1 in rbps:
            csv_file.write(rbp1 + ",")
            for rbp2 in rbps:
                csv_file.write(str(symmetric_corr_table[rbp1][rbp2]))
                csv_file.write(",")
            csv_file.write("\n")

    return path_to_save


def generate_heat_map(_, rna_info, __, ___):
    """
    Function meant to generate a heatmap using correlation values.
    Unimplemented.
    :param _:
    :param rna_info:
    :param __:
    :param ___:

    """
    del rna_info
    print("HEAT MAP (supposed to be) GENERATED")


def overall_correlation_analysis(big_storage, rna_info, configs=None):
    """
    A data analysis function (designed for the CLI variant of RNPFind) that
    allows one to calculate and generate CSV files and/or heatmaps for binding
    correlation between RBPs binding on an RNA molecule of interest.

    :param big_storage: A dictionary keyed-in by data load source name. Each
        such data load source key should correspond to a Storage instance
        containing binding sites extracted from said data source (on the RNA of
        interest).
    :param rna_info: Dictionary containing key "official_name", corresponding
        to the name of the RNA molecule of interest.
    :param config: A dictionary of additional paramters. The following keys are
        useful:
         - 'base_stringency': specifies number of bases to consider to be
           competing.
         - 'out_dir': specifies the directory to write csv files to.

    """

    threshold = configs["base_stringency"]

    symmetric_corr_tables = {}
    data_load_sources = big_storage.keys()
    for data_load_source in data_load_sources:
        # print("")
        # print("Working on ", data_load_source, "data...")

        storage = big_storage[data_load_source]

        _, symmetric_corr_table, _, _ = storage.self_analysis(
            bp_threshold=threshold, progress_feedback=False
        )
        symmetric_corr_tables[data_load_source] = symmetric_corr_table
        # print("Done!")

    # print("Creating CSV files...")
    for data_load_source in data_load_sources:
        symmetric_corr_table = symmetric_corr_tables[data_load_source]
        path_saved = generate_csv(
            symmetric_corr_table,
            rna_info,
            [data_load_source],
            threshold,
            configs["out_dir"],
        )
        print("A file was saved at", path_saved + ".", file=sys.stderr)
    # print("Done!")



"""
Tests the Attract data loading functions for correctness

"""
import unittest

from src.rnpfind.analysis_functions import (
    analysis_method_functions,
    analysis_methods_supported_long,
    analysis_methods_supported_short,
)

# from .attract_data_load import attract_data_load
from src.rnpfind.gene_coordinates import Chromosome


class TestCorrectBindingSites(unittest.TestCase):
    """
    Tests to ensure that the ATTRACT data load function loads binding sites
    that are consistent with that of the web interface.

    """

    def test_correct_sites(self):
        """
        Checks that the binding sites match with web interface.

        """

        test_rna_info = {}
        test_rna_info["official_name"] = "Malat1"
        test_rna_info["chr_n"] = Chromosome(11)
        test_rna_info["start_coord"] = 65497738
        test_rna_info["end_coord"] = 65506516

        # attract_binding_sites = attract_data_load(test_rna_info)
        self.assertEqual(
            len(analysis_methods_supported_short),
            len(analysis_methods_supported_long),
        )
        self.assertEqual(
            len(analysis_method_functions),
            len(analysis_methods_supported_short),
        )


if __name__ == "__main__":
    unittest.main()



"""
Defines functions that are involved in getting some data from the user.
Used for the command line interface variant of RNPFind.
"""
import sys

from hgfind import WrongGeneName, hgfind

from .analysis_functions import (
    analysis_methods_supported_long,
    analysis_methods_supported_short,
)
from .config import GENOME_VERSION
from .data_load_functions import (
    data_load_sources_supported_long,
    data_load_sources_supported_short,
)
from .gene_coordinates import Chromosome


def parse_genome_coord(transcript: str) -> dict:
    """
    Given a string of the form <chr>:<start>-<end> representing genomic
    coordinates, extracts each of the chromosome number, start, and end
    coordinates.

    The input may be malformed.

    :param transcript: input genomic coordinates
    :returns: a dictionary containing the following keys and associated values:
        'success': whether the conversion worked or not. If this is False, the
            other keys below are not guaranteed to exist.
        'chr_n': the Chromosome object on which the gene lies
        'start_coord': the start coordinate of the gene on the chromosome
        'end_coord': the end coordinate of the gene on the chromosome
    """

    fail_output = {"success": False}

    # import pdb
    # pdb.set_trace()

    # Check if structure seems correct
    if len([c for c in transcript if c == ":"]) != 1:
        return fail_output
    if len([c for c in transcript if c == "-"]) != 1:
        return fail_output
    if transcript.find(":") > transcript.find("-"):
        return fail_output

    chr_n = transcript.split(":")[0]
    try:
        chr_n = Chromosome(chr_n)
        start_coord = transcript.split(":")[1].split("-")[0]
        end_coord = transcript.split(":")[1].split("-")[1]
        start_coord = int(start_coord)
        end_coord = int(end_coord)
    except ValueError:
        return fail_output

    if start_coord > end_coord:
        return fail_output

    success_output = {
        "success": True,
        "chr_n": chr_n,
        "start_coord": start_coord,
        "end_coord": end_coord,
    }

    return success_output


def get_rna_coord(rna_gene_name):
    """
    Given a gene name, informs the user of the predicted RNA coordinates.
    Allows the user to potentially revise the coordinates, if required.

    :param rna_gene_name: input RNA gene name (string)

    """

    try:
        rna_info = hgfind(rna_gene_name)
    except WrongGeneName:
        # Transcript might have been given in <chr no>:<start>-<end> form
        rna_info = parse_genome_coord(rna_gene_name)
        if not rna_info["success"]:
            print(
                "Error parsing transcript. Please give a correct gene name"
                "or a genomic coordinate of the form <chr>:<start>-<end>",
                file=sys.stderr,
            )
            sys.exit(1)

    rna_chr_no = rna_info["chr_n"]
    rna_start_chr_coord = rna_info["start_coord"]
    rna_end_chr_coord = rna_info["end_coord"]

    print(
        f"Analyzing {GENOME_VERSION}"
        f" {rna_chr_no}:{rna_start_chr_coord}-{rna_end_chr_coord}"
        f" (length = {rna_end_chr_coord - rna_start_chr_coord} bases)",
        file=sys.stderr,
    )

    return rna_chr_no, rna_start_chr_coord, rna_end_chr_coord


def get_user_rna_preference(transcript: str) -> dict:
    """
    Asks the user for the RNA molecule they wish to analyze (the name of the
    gene is expected). Converts the gene name into coordinates.

    :returns: a dictionary containing the official name of the RNA, along with
              its chrosome number, start coordinate, and end coordinate on the
              human genome.
    """
    rna_chr_no, rna_start_chr_coord, rna_end_chr_coord = get_rna_coord(
        transcript
    )
    rna_info = {}
    rna_info["official_name"] = (
        transcript.upper()
        if ":" not in transcript
        else f"chr{rna_chr_no}-gene"
    )
    rna_info["chr_n"] = rna_chr_no
    rna_info["start_coord"] = int(rna_start_chr_coord)
    rna_info["end_coord"] = int(rna_end_chr_coord)
    return rna_info


def get_user_data_source_preference():
    """
    Asks the user for their preferred method of data source retrieval.
    The user is asked to input numbers to represent their combination of
    requested data sources
    """
    print("")
    print("")
    print(
        "Which sources of data would you like to collect RBP binding data"
        " from today?"
    )
    print("")
    for i, source in enumerate(data_load_sources_supported_long):
        print("[" + str(i) + "]: " + source)
    print("")

    input_digits = ""
    while not (
        input_digits.isdigit()
        and all(
            [
                int(c) < len(data_load_sources_supported_short)
                for c in input_digits
            ]
        )
    ):
        print(
            "Please choose any combination from above as you like (e.g. 124)"
        )
        print(">")
        input_digits = input()

    print("Thank you")
    return [data_load_sources_supported_short[int(i)] for i in input_digits]


def get_user_analysis_preference():
    """
    Asks the user for their preferred method of data analysis on retrieved
    binding sites. The user is asked to input numbers to represent their
    combination of requested data sources.
    """
    print("")
    print("")
    print(
        "Which method of analysis on the data would you like to employ today?"
    )
    print("")
    for i, source in enumerate(analysis_methods_supported_long):
        print("[" + str(i) + "]: " + source)
    print("")
    print(
        "Please pick just one analysis method and write the number associated"
        " with it (e.g. 3)"
    )
    print(">")
    input_digits = input()
    while not (
        len(input_digits) == 1
        and input_digits.isdigit()
        and int(input_digits) < len(analysis_methods_supported_short)
    ):
        print(
            "Please pick just one analysis method and write the number"
            " associated with it (e.g. 3)"
        )
        print(">")
        input_digits = input()

    print("Thank you")
    return analysis_methods_supported_short[int(input_digits)]



"""
This module defines functions relevant to loading RBP binding sites from the
ATTRACT database.

"""

from .config import ANNOTATION_COLUMN_DELIMITER, ATTRACT_PATH
from .picklify import picklify
from .pwm_scan import get_human_seq, pwm_scan, str_to_pwm

attract_all_column_names = [
    "Gene_name",
    "Gene_id",
    "Mutated",
    "Organism",
    "Motif",
    "Len",
    "Experiment_description",
    "Database",
    "Pubmed",
    "Experiment",
    "Family",
    "Matrix_id",
    "attractScore",
]
attract_all_column_descriptions = attract_all_column_names
attract_columns_of_interest = [2, 4, 6, 7, 8, 9, 10, 11, 12]
attract_default_label_index = [8]
ATTRACT_DEFAULT_MOUSE_OVER_INDEX = 9

attract_column_names = [
    attract_all_column_names[i] for i in attract_columns_of_interest
]
attract_column_descriptions = [
    attract_all_column_descriptions[i] for i in attract_columns_of_interest
]


def generate_matrix_to_pwm_dict():
    """
    Preprocesses ATTRACT database files. In particular, this function generates
    and returns a dictionary that maps position weight matrix IDs to the PWM
    data structures used in this program (RNPFind).
    """
    attract_pwm_file_path = f"{ATTRACT_PATH}/attract-pwm.txt"
    matrix_to_pwm_dict = {}
    with open(attract_pwm_file_path) as handle:
        line = handle.readline()
        while line:
            matrix_id = line.split()[0][1:]
            raw_pwm_str = ""
            line = handle.readline()
            while line and line[0] != ">":
                raw_pwm_str += line
                line = handle.readline()
            # Now we have the raw text, we convert it to pwm and add to
            # dictionary
            matrix_to_pwm_dict[matrix_id] = str_to_pwm(raw_pwm_str)
    return matrix_to_pwm_dict


def attract_data_load(rna_info):
    """
    Loads RBP binding sites from the ATTRACT database for a given RNA molecule,
    and returns the sites as a Generator / Iterator object.
    :param rna_info: a dictionary containing the location of the RNA on the
        hg38 genome by specifying its chromosome location, start coordinates,
        and end coordinates.

    """
    attract_protein_file_path = f"{ATTRACT_PATH}/ATtRACT_db.txt"
    rna_seq = get_human_seq(rna_info)
    matrix_to_pwm_dict = picklify(generate_matrix_to_pwm_dict)
    with open(attract_protein_file_path) as handle:
        columns = handle.readline().strip().split("\t")
        assert columns == [
            "Gene_name",
            "Gene_id",
            "Mutated",
            "Organism",
            "Motif",
            "Len",
            "Experiment_description",
            "Database",
            "Pubmed",
            "Experiment_description",
            "Family",
            "Matrix_id",
            "Score",
        ]
        protein_columns = handle.readline().replace("\n", "").split("\t")
        while protein_columns != [""]:
            # Warning: Score ends with \n here, maybe remove using strip or
            # indexing. For now, we don't care about score as it seems to be
            # about literature

            # We only care about human RBPs for now.
            if protein_columns[3] != "Homo_sapiens":
                protein_columns = (
                    handle.readline().replace("\n", "").split("\t")
                )
                continue

            annotation = ANNOTATION_COLUMN_DELIMITER.join(
                [protein_columns[i] for i in attract_columns_of_interest]
            )

            rbp = protein_columns[0]

            matrix_id = protein_columns[11]

            pwm = matrix_to_pwm_dict[matrix_id]
            sites = pwm_scan(rna_seq, pwm)
            if not sites:
                protein_columns = (
                    handle.readline().replace("\n", "").split("\t")
                )
                continue

            for start, end in sites:
                yield rbp, start, end, annotation

            protein_columns = handle.readline().replace("\n", "").split("\t")



"""
This file is responsible for managing all the data loading functions that
RNPFind supports. If you want to develop and incorporate your own data loading
function into RNPFind, this is the file to edit!

The way any data loading function works in RNPFind is as follows. RNPFind will
call the data loading function like this: data_load_function_name(rna_info)

where rna_info is a list that stores the RNA name and its chromosomal location
on the hg38 chromosome. If you preferred to have the RNA sequence,
get_human_seq() from pwm_scan should help you (see attract_data_load.py for an
example of that).

Upon getting the above input, a data loading function must return an iterable
object that contains elements of the following form that each represent a
binding site of an rbp on the RNA molecule: (rbp, start, end, annotation). For
example, the function might return a list that looks like
[
    (rbp_1, start_1, end_1, annotation_1),
    (rbp_2, start_2, end_2, annotation_2), ...
    (rbp_n, star_n, end_n, annotation_n)
].

See attract_data_load.py for an example where "yield" was used to make an
iterator instead that serves the same purpose (more out of a laziness of
changing code than any demonstrable efficiency).

Above, for each binding site, "rbp" is the name of an RBP molecule (a string).

The "start" refers to the starting coordinate of a binding site relative to the
start of the RNA (where "start" is always the thing on the left side, doesn't
matter if the gene is on anti-sense strand) on the hg38 chromosome, and "end"
refers to the nucleotide number where the site ends (again relative to where the
RNA starts on the chromosome).

"Annotation" is any annotation you might want to add to the binding site. I have
struggled with standardizing this, switching back and forth between strings with
delimiters or just a list of annotations. For now, I have stuck with the
following standardization: the data loading function should give a list of
"features" that it wants to associate with the binding site. For example, a
binding site may be associated with the features "database id", "binding score",
and "experiment_type", so the function would return something like
(rbp_1, start_1, end_1, ["nm00042", 0.9, "clip seq"]). Internally, RNPFind does
stuff like converting it to a string separated by delimiters, but that's just a
relic from the past and my lack of firm will in changing the code.

Above, annotation to the binding sites are given just as data, but what about
the "column headings", so to speak? That's exactly a part of what this file
does!

If you make a data loading function to the above specification, the below lines
should be edited to incorporate the function on RNPFind!
"""

from .attract_data_load import (
    ATTRACT_DEFAULT_MOUSE_OVER_INDEX,
    attract_column_descriptions,
    attract_column_names,
    attract_data_load,
    attract_default_label_index,
)

# These imports are more obvious once you see below
from .colors import blue, dark_green, orange
from .custom_data_load import custom_data_load
from .postar_data_load import (
    POSTAR_DEFAULT_MOUSE_OVER_INDEX,
    postar_column_descriptions,
    postar_column_names,
    postar_data_load,
    postar_default_label_index,
)
from .rbpdb_data_load import (
    RBPDB_DEFAULT_MOUSE_OVER_INDEX,
    rbpdb_column_descriptions,
    rbpdb_column_names,
    rbpdb_data_load,
    rbpdb_default_label_index,
)

# First, add a short name for your data source method to this list
data_load_sources_supported_short = ["rbpdb", "attract", "postar"]

# Now give it a long name
data_load_sources_supported_long = [
    "RBPDB (computational)",
    "ATTRACT (computational)",
    "POSTAR (experimental)",
]

# Link your short name to the actual data loading function that complies with
# the above specifications here.

# This is where importing is useful!
data_load_sources_functions = {
    "rbpdb": rbpdb_data_load,
    "attract": attract_data_load,
    "postar": postar_data_load,
    "custom": custom_data_load,
}

# Finally, for the sake of UCSC visualization, give a little bit info about
# your data source function's annotations. Map the short name of your data load
# function to another dictionary, which has the following mappings:
#   "names": mapped to a list of column headings (strings) that the annotations
#       have
#   "default_label": mapped to the index of the column heading that would be
#       preferred to be displayed next to each binding site on the UCSC view.
#       This feature is NOT implemented yet.
#   "default_mouse_over": mapped to the index of the column heading that would
#       be preferred to be displayed when a mouse is hovered on a binding site
#       on the UCSC view. This feature is NOT implemented yet.
#   "descriptions: mapped to a list of descriptions for each of the column
#       headings. This could be the same as names, if the names are
#       self-explanatory.

column_data = {
    "postar": {
        "names": postar_column_names,
        "default_label": postar_default_label_index,
        "default_mouse_over": POSTAR_DEFAULT_MOUSE_OVER_INDEX,
        "descriptions": postar_column_descriptions,
    },
    "attract": {
        "names": attract_column_names,
        "default_label": attract_default_label_index,
        "default_mouse_over": ATTRACT_DEFAULT_MOUSE_OVER_INDEX,
        "descriptions": attract_column_descriptions,
    },
    "rbpdb": {
        "names": rbpdb_column_names,
        "default_label": rbpdb_default_label_index,
        "default_mouse_over": RBPDB_DEFAULT_MOUSE_OVER_INDEX,
        "descriptions": rbpdb_column_descriptions,
    },
}

# When the data source is viewed on UCSC, what color should the binding sites
# be? Give a tuple with three numbers, indicating the rgb values of the color
# (each ranging from 0 to 255, as is common).

data_load_source_colors = {
    "postar": orange,
    "attract": blue,
    "rbpdb": dark_green,
}

# To consider: right now, I have mainly filtered through the "columns of
# interest" for each of the data loading functions WITHIN the data loading
# functions, just to be selective about which columns are interesting to view
# on the UCSC visualization feature. This seems like bad practice; maybe in the
# future someone will develop more analysis functions that are interested in
# some of the other columns of the data; it might be wise to consider loading
# all columns of data at the data loading level and letting the analysis
# functions individually deal with which of the columns they are interested in.



"""
This module is dedicated to functions that help with converting a string
representing a gene name to its coordinates on the hg38 chromosome.

"""

from typing import Union


def is_int(in_obj):
    """
    Checks if the input represents an integer and returns true iff so

    """
    try:
        int(in_obj)
        return True
    except ValueError:
        return False


class Chromosome:
    """
    This class allows one to specify chromsomes from the human genome.
    Benefits include comparison of chromosomes that are sex chromosomes,
    autosomes, or genes included in the mitochondrial DNA.

    """

    def __init__(self, n: Union[int, str]):

        if isinstance(n, int) and 1 <= n <= 22:
            self.chr_n = n
        elif isinstance(n, str) and n.upper() in ("X", "Y", "MT", "M"):
            if n.upper() == "X":
                self.chr_n = 23
            elif n.upper() == "Y":
                self.chr_n = 24
            elif n.upper() == "MT":
                self.chr_n = 25
            elif n.upper() == "M":
                self.chr_n = 25
        elif isinstance(n, str) and is_int(n) and 1 <= int(n) <= 22:
            self.chr_n = int(n)
        else:
            raise ValueError("expected X, Y, M(T), or [1-22]")

    def __str__(self):
        return (
            str(self.chr_n)
            if self.chr_n <= 22
            else "X"
            if self.chr_n == 23
            else "Y"
            if self.chr_n == 24
            else "M"
        )

    def __lt__(self, other):
        return self.chr_n < other.chr_n

    def __eq__(self, other):
        return self.chr_n == other.chr_n

    def __hash__(self):
        return hash(self.chr_n)

    def __gt__(self, other):
        return self.chr_n > other.chr_n

    def __repr__(self):
        return self.__str__()

    def __int__(self):
        return self.chr_n


if __name__ == "__main__":
    pass



"""
This file is responsible for managing all the analysis functions that RNPFind
supports. If you want to develop and incorporate your own analysis function into
RNPFind, this is the file to edit!

The way any analysis function works in RNPFind is as follows. RNPFind will call
the analysis function like this:

analysis_method_function_name(big_storage, rna_info)

where big_storage is a dictionary that maps a data_source string to a Storage
variable (which in turn maps rbp names to a list of binding sites; for more
details see load_data.py), and rna_info is a dictionary that stores the RNA name
and its chromosomal location on the hg38 chromosome. If you preferred to have
the RNA sequence, get_human_seq() from pwm_scan should help you (see
attract_data_load.py for an example of that).

Upon getting the above input, an analysis function is free to do with the
information as it pleases. The information it outputs to the user could be in
any form, such as text on the CLI, csv files saved on the local directory, heat
maps produced on a figure generated, or a URL to web-generated content. Once
done with the work, the function should return so that RNPFind can continue with
its process.

If an analysis function is made to the above specification, the below lines
should be edited to incorporate the function on RNPFind!
# """

# First, import the function from wherever it is defined:
from .over_all_corr_analysis import overall_correlation_analysis
from .ucsc_visualize import ucsc_visualize

# Give your method a short name:
analysis_methods_supported_short = ["csv", "bed"]

# Give your method a long name:
analysis_methods_supported_long = [
    "Binding correlation csv",
    "Bed files with binding data",
]

# Map your short name to the variable imported above that corresponds to your
# function!
analysis_method_functions = {
    "csv": overall_correlation_analysis,
    "bed": ucsc_visualize,
}
